@@ -317,7 +317,7 @@ keywords: else, elif, fi, done, when, case.
   "Return true if the point is at the beginning of a defun"
   (or (looking-at "[ \t]*prog [a-z]+")
       (looking-at (concat "[ \t]*"
-			  (regexp-opt '("begin" "end") 'words)))
+			  (regexp-opt '("startup" "shutdown" "begin" "end") 'words)))
       (looking-at "[ \t]*func\\s +[a-zA-Z_][a-zA-Z0-9_]*\\s *(")))
 
 (defun mfl-search-next-defun ()
@@ -355,15 +355,15 @@ otherwise."
 
 (defconst mfl-keywords
   (regexp-opt '("accept" "add" "and" "alias" "begin" "break" "bye" "case"
-	        "catch" "const" "continue" "default"
+	        "catch" "const" "continue" "dclex" "default"
 		"delete" "discard" "do" "done"
 		"echo" "end" "elif" "else"
 		"fi" "fnmatches" "for" "from" "func"
 		"if" "import" "loop" "matches" "module" "next"
 		"not" "on" "or" "pass" "precious"
 		"prog" "reject" "replace" "return"
-		"returns" "require" "set" "static" "switch" "tempfail"
-		"throw" "try" "vaptr" "when" "while") 'words))
+		"returns" "require" "set" "shutdown" "startup" "static" "switch"
+		"tempfail" "throw" "try" "vaptr" "when" "while") 'words))
 
 (defconst mfl-on-keywords
   ;; context-dependent keywords
