@@ -48,7 +48,7 @@ dns_log_cb(adns_state ads, void *logfndata, const char *fmt, va_list al)
    but it will emit \e<N> directives in the middle of the string, which
    upsets the mailutils' logstream implementation.
 
-   A possible work over would be to use logfndata to select between
+   A possible workaround would be to use logfndata to select between
    mu_diag_vprintf,mu_diag_cont_vprintf or appropriate mu_debug_log_
    call.
 
@@ -56,11 +56,13 @@ dns_log_cb(adns_state ads, void *logfndata, const char *fmt, va_list al)
 	mu_stream_vprintf(mu_strerr, fmt, al);
 }
 
-void
-dnsbase_init(void)
+static void
+dnsbase_finish(void)
 {
-	if (!debug_handle)
-		debug_handle = mu_debug_register_category("dns");
+	if (state) {
+		adns_finish(state);
+		state = NULL;
+	}
 }
 
 void
@@ -69,7 +71,11 @@ dnsbase_real_init(char *configtext)
 	int rc;
 	int flags;
 	mu_debug_level_t lev;
-	
+	static int cleanup_installed;
+
+	/* Dispose of a previosly allocated state, if any */
+	dnsbase_finish();
+
 	flags = adns_if_nosigpipe;
 	if (mu_debug_get_category_level(debug_handle, &lev) == 0
 	    && (lev & MU_DEBUG_LEVEL_MASK(MU_DEBUG_TRACE9)))
@@ -79,10 +85,14 @@ dnsbase_real_init(char *configtext)
 		mu_diag_funcall(MU_DIAG_ERROR, "adns_init", NULL, rc);
 		exit(1);
 	}
+	if (!cleanup_installed) {
+		atexit(dnsbase_finish);
+		cleanup_installed = 1;
+	}
 }
 
 void
-dnsbase_file_init(char *filename)
+dnsbase_file_init(char const *filename)
 {
 	if (!filename)
 		dnsbase_real_init(NULL);
@@ -263,7 +273,7 @@ int adns_to_dns_tab[] = {
 	[adns_s_unknownrrtype]       = STAT(dns_failure),
 	[adns_s_systemfail]          = STAT(dns_failure),
 
-	/* remotely induced errors), detected locally */
+	/* remotely induced errors, detected locally */
 	[adns_s_timeout]             = STAT(dns_temp_failure),
 	[adns_s_allservfail]         = STAT(dns_temp_failure),
 	[adns_s_norecurse]           = STAT(dns_temp_failure),
@@ -327,6 +337,161 @@ adns_to_dns_status(int e)
 	return dns_not_found;
 }
 
+/*
+ * dns_query and friends - a query wrapper.
+ *
+ * Adns library expressly disallows CNAMEs pointing to another
+ * CNAMEs, and a good thing it does (for the reference, see RFC 1034,
+ * section 3.6.2).  However, reportedly such CNAME chains are being
+ * used quite often, e.g. for pointing to TXT records.  Thus, a practical
+ * need for supporting CNAME chains to some extent does exist.
+ *
+ * The dns_query function below is a wrapper over adns_synchronous that
+ * follows CNAME with limited length, with CNAME loop detection.  The
+ * maximum length of a chain is given by the dns_max_cname_chain variable.
+ * Values 0 and 1 disable CNAME chain support.  If a the length limit is
+ * hit or a CNAME loop is detected, adns_s_prohibitedcname status is
+ * returned.
+ */
+size_t dns_max_cname_chain = 2;
+
+/*
+ * To track encountered CNAMES, a singly linked list is used.  This
+ * means, among others, that dns_max_cname_chain better be sufficiently
+ * small.
+ */
+struct cname_record {
+	struct cname_record *next; /* Pointer to next record */
+	char name[1];              /* Actual name follows the structure */
+};
+
+/* A list of recorded CNAMEs */
+struct cname_record_list {
+	size_t count;              /* Number of elements in the list. */
+	struct cname_record *head, *tail;
+};
+
+#define CNAME_RECORD_LIST_INITIALIZER { 0, NULL, NULL }
+
+/*
+ * Free the list entries from RECLIST.  The structure RECLIST points to
+ * should be allocated on stack, and therefore not freed.
+ */
+static void
+cname_record_list_free(struct cname_record_list *reclist)
+{
+	struct cname_record *rec = reclist->head;
+	while (rec) {
+		struct cname_record *next = rec->next;
+		free(rec);
+		rec = next;
+	}
+}
+
+/*
+ * Install NAME into RECLIST.  Return pointer to the allocated copy
+ * of the name.  Return NULL if NAME is already stored in the list.
+ */
+static char const *
+cname_install(struct cname_record_list *reclist, char const *name)
+{
+	struct cname_record *rec;
+
+	for (rec = reclist->head; rec; rec = rec->next) {
+		if (mu_c_strcasecmp(rec->name, name) == 0)
+			return NULL;
+	}
+	rec = mu_alloc(sizeof(*rec) + strlen(name));
+	strcpy(rec->name, name);
+	rec->next = NULL;
+	if (!reclist->head)
+		reclist->head = rec;
+	else
+		reclist->tail->next = rec;
+	reclist->tail = rec;
+	reclist->count++;
+	return rec->name;
+}
+
+/*
+ * dns_query - look up a label NAME of RR type TYPE in the DNS.  Follow
+ * CNAME chains of up to dns_max_cname_chain elements.  In other respects
+ * the behavior is the same as that of adns_synchronous.
+ *
+ * FIXME: in the presence of a CNAME chain, this function does two
+ * extra lookups, compared with the hypothetical libresolv implementation.
+ * This is due to the specifics of libadns.
+ */
+int
+dns_query(const char *name, adns_rrtype type, adns_answer **ans_ret)
+{
+	adns_state state = get_state();
+	adns_answer *ans = NULL, *cnans = NULL;
+	int rc;
+
+	/*
+	 * First, look up the requested RR type.  If the actual record is
+	 * a CNAME pointing to the requested RR, this will be handled by
+	 * adns due to adns_qf_cname_loose flag in DEFAULT_QFLAGS.
+	 *
+	 * If it is a CNAME pointing to a CNAME, this will result in the
+	 * first extra lookup (see FIXME above).
+	 */
+	rc = adns_synchronous(state, name, type, DEFAULT_QFLAGS, &ans);
+	if (rc == 0 && ans->status == adns_s_prohibitedcname
+	    && dns_max_cname_chain > 1) {
+		struct cname_record_list cname_rec = CNAME_RECORD_LIST_INITIALIZER;
+
+                /* Record the queried name, first. */
+		cname_install(&cname_rec, name);
+
+		/* Follow the CNAME chain. */
+		while (cname_rec.count - 1 <= dns_max_cname_chain) {
+			if (adns_synchronous(state, name, adns_r_cname,
+					     DEFAULT_QFLAGS, &cnans))
+				break;
+			if (cnans->status == adns_s_ok) {
+				/*
+				 * CNAME found. Record it and continue.
+				 */
+				name = cname_install(&cname_rec, cnans->rrs.str[0]);
+				free(cnans);
+				if (!name)
+					/*
+					 * Loop detected.  Returned ans
+					 * retains the adns_s_prohibitedcname
+					 * status.
+					 */
+					break;
+			} else if (cnans->status == adns_s_nodata) {
+				/*
+				 * RR found, but has a different type.
+				 * Look up the requested type using the last
+				 * recorded name.  This accounts for second
+				 * extra lookup.
+				 */
+				free(cnans);
+				rc = adns_synchronous(state, name, type, DEFAULT_QFLAGS, &ans);
+				break;
+			} else {
+				/*
+				 * Another error.  Replace original answer with
+				 * the last one.
+				 */
+				free(ans);
+				ans = cnans;
+			}
+		}
+		cname_record_list_free(&cname_rec);
+	}
+
+	if (rc == adns_s_ok)
+		*ans_ret = ans;
+	else
+		free(ans);
+	return rc;
+}
+
 dns_status
 soa_check(const char *name, int ip, struct dns_reply *reply)
 {
@@ -334,9 +499,7 @@ soa_check(const char *name, int ip, struct dns_reply *reply)
 	int rc;
 	adns_answer *ans;
 
-	rc = adns_synchronous(get_state(), name, adns_r_soa_raw,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(name, adns_r_soa_raw, &ans);
 	if (rc)
 		return errno_to_dns_status(rc);
 	status = adns_to_dns_status(ans->status);
@@ -386,9 +549,7 @@ mx_lookup(const char *host, int resolve, struct dns_reply *reply)
 	adns_answer *ans;
 	int i;
 
-	rc = adns_synchronous(get_state(), host, adns_r_mx,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(host, adns_r_mx, &ans);
 	if (rc)
 		return errno_to_dns_status(rc);
 	status = adns_to_dns_status(ans->status);
@@ -456,9 +617,7 @@ dns_resolve_ipstr(const char *ipstr, const char *domain, char **hbuf)
 		type = adns_r_a;
         }
 
-	rc = adns_synchronous(get_state(), name, type,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(name, type, &ans);
 	free(name);
 	if (rc)
 		return errno_to_dns_status(rc);
@@ -481,9 +640,7 @@ dns_resolve_hostname(const char *host, char **ipbuf)
 	int rc;
 	adns_answer *ans;
 	
-	rc = adns_synchronous(get_state(), host, adns_r_a,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(host, adns_r_a, &ans);
 	if (rc)
 		return errno_to_dns_status(rc);
 	status = adns_to_dns_status(ans->status);
@@ -501,9 +658,7 @@ a_lookup(const char *host, struct dns_reply *reply)
 	int rc;
 	adns_answer *ans;
 	
-	rc = adns_synchronous(get_state(), host, adns_r_a,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(host, adns_r_a, &ans);
 	if (rc)
 		return errno_to_dns_status(rc);
 	status = adns_to_dns_status(ans->status);
@@ -531,9 +686,7 @@ ptr_lookup(struct in_addr ip, struct dns_reply *reply)
 		    (ip.s_addr >> 8) & 0xff,
 		    (ip.s_addr >> 16) & 0xff,
 		    (ip.s_addr >> 24) & 0xff);
-	rc = adns_synchronous(get_state(), name, adns_r_ptr_raw,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(name, adns_r_ptr_raw, &ans);
 	free(name);
 	if (rc)
 		return errno_to_dns_status(rc);
@@ -555,9 +708,7 @@ txt_lookup(const char *name, struct dns_reply *reply)
 	int rc;
 	adns_answer *ans;
 
-	rc = adns_synchronous(get_state(), name, adns_r_txt,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(name, adns_r_txt, &ans);
 	if (rc)
 		return errno_to_dns_status(rc);
 	status = adns_to_dns_status(ans->status);
@@ -769,9 +920,7 @@ ns_lookup(const char *domain, int resolve, struct dns_reply *reply)
 	adns_answer *ans;
 	int i;
 
-	rc = adns_synchronous(get_state(), domain, adns_r_ns_raw,
-			      DEFAULT_QFLAGS,
-			      &ans);
+	rc = dns_query(domain, adns_r_ns_raw, &ans);
 	if (rc)
 		return errno_to_dns_status(rc);
 	status = adns_to_dns_status(ans->status);
@@ -788,3 +937,38 @@ ns_lookup(const char *domain, int resolve, struct dns_reply *reply)
 	
 	return status;
 }
+
+int
+cb_resolv_conf(void *data, mu_config_value_t *arg)
+{
+	if (mu_cfg_assert_value_type(arg, MU_CFG_STRING))
+		return 1;
+	dnsbase_file_init(arg->v.string);
+	return 0;
+}
+
+struct mu_cfg_param resolver_section_param[] = {
+	{ "config", mu_cfg_callback,
+	  NULL, 0, cb_resolv_conf,
+	  N_("Read DNS configuration from this file (default: /etc/resolv.conf)"),
+          N_("name: string") },
+	{ "max-cname-chain", mu_c_size,
+	  &dns_max_cname_chain, 0, NULL,
+	  N_("Maximum allowed length of CNAME chains") },
+	{ NULL }
+};
+
+void
+dnsbase_init(void)
+{
+	struct mu_cfg_section *section;
+	if (!debug_handle)
+		debug_handle = mu_debug_register_category("dns");
+	if (mu_create_canned_section("resolver", &section) == 0) {
+		section->label = NULL;
+		section->parser = NULL;
+		section->docstring = N_("DNS resolver configuration.");
+		mu_cfg_section_add_params(section, resolver_section_param);
+	}
+}
+
