@@ -20,7 +20,7 @@
 #define DEFAULT_RATE_DATABASE "rates.db"
 #define DEFAULT_GREYLIST_DATABASE "greylist.db"
 #define DEFAULT_DNS_DATABASE "dns.db"
-#define DEFAULT_SCRIPT_FILE SYSCONFDIR "/mailfromd.mf"
+#define DEFAULT_SCRIPT_FILE_STEM "mailfromd"
 #define DEFAULT_CONFIG_FILE SYSCONFDIR "/mailfromd.conf"
 #define DEFAULT_FROM_ADDRESS "<>"
 #ifdef DEFAULT_PREPROCESSOR
