@@ -140,7 +140,9 @@ int switch_to_privs (uid_t uid, gid_t gid, struct mf_gid_list *);
 void get_user_groups (struct mf_gid_list *grouplist, const char *user);
 
 void mf_priv_setup (struct mf_privs *);
-void mf_epriv_setup (struct mf_privs *);
+
+extern void (*mf_server_startup_hook)(void);
+extern void (*mf_server_shutdown_hook)(void);
 
 
 char **config_array_to_argv (mu_config_value_t *val);
@@ -149,8 +151,6 @@ char *config_array_to_string (mu_config_value_t *val);
 int config_cb_timeout (struct timeval *pt, mu_config_value_t *val);
 int config_cb_time_t(void *data, mu_config_value_t *arg);
 int config_cb_ignore(void *data, mu_config_value_t *val);
-int config_cb_lock_retry_count(void *data, mu_config_value_t *val);
-int config_cb_lock_retry_timeout(void *data, mu_config_value_t *val);
 
 void mf_proctitle_init (int argc, char *argv[], char *env[]);
 void mf_proctitle_format (const char *fmt, ...);
@@ -241,6 +241,7 @@ void logger_open(void);
 void logger_close(void);
 void logger_fdset(fd_set *set);
 void logger_text(int prio, const char *text);
+void logger_set_print_severity(int enable);
 
 #define vlogmsg mu_diag_voutput
 #define logmsg mu_diag_output
