@@ -179,11 +179,6 @@ logger_open()
 	
 	if (lp->log_open)
 		lp->log_open();
-	mu_stream_ioctl(mu_strerr, MU_IOCTL_LOGSTREAM,
-			MU_IOCTL_LOGSTREAM_GET_MODE, &mode);
-	mode |= MU_LOGMODE_SEVERITY;
-	mu_stream_ioctl(mu_strerr, MU_IOCTL_LOGSTREAM,
-			MU_IOCTL_LOGSTREAM_SET_MODE, &mode);
 	mode = MU_DEBUG_LEVEL_MASK(MU_LOG_ERROR);
 	mu_stream_ioctl(mu_strerr, MU_IOCTL_LOGSTREAM,
 			MU_IOCTL_LOGSTREAM_SET_SEVERITY_MASK, &mode);
@@ -213,4 +208,10 @@ logger_text(int prio, const char *text)
 	(lp->log_text ? lp->log_text : syslog_default_logtext)(prio, text);
 }
 
-		
+void
+logger_set_print_severity(int enable)
+{
+	mu_log_print_severity = enable;
+	mu_stream_ioctl(mu_strerr, MU_IOCTL_LOGSTREAM,
+			MU_IOCTL_LOGSTREAM_SET_MODE, &enable);
+}
