@@ -136,6 +136,8 @@ struct mf_gid_list *mf_server_retain_groups;
                                      /* List of group IDs to retain when
 					switching to user privileges. */
 
+void (*mf_server_startup_hook)(void);
+void (*mf_server_shutdown_hook)(void);
 
 void
 priv_setup()
@@ -269,6 +271,8 @@ mf_server_start(const char *program, const char *dir, const char *pidfile,
 		umask(mf_server_umask);
 	}
 
+	if (mf_server_startup_hook)
+		mf_server_startup_hook();
 	mf_server_log_setup();
 
 	mu_error(_("%s started"), program);
@@ -280,12 +284,13 @@ mf_server_start(const char *program, const char *dir, const char *pidfile,
 	mfd_srvman_shutdown();
 	mfd_srvman_free();
 	unlink(pidfile);
+	if (mf_server_shutdown_hook)
+		mf_server_shutdown_hook();
 	if (restart) {
 		char *s;
 		
 		mu_error(_("%s restarting"), program);
 		mu_daemon_remove_pidfile();
-		mu_onexit_run();
 		logger_close();
 		if (logger_flags(LOGF_STDERR))
 			close_fds_above(2);
@@ -298,5 +303,6 @@ mf_server_start(const char *program, const char *dir, const char *pidfile,
 			mu_error(_("command line was: %s"), s);
 	} else 
 		mu_error(_("%s terminating"), program);
+
 	exit(0);
 }
