@@ -300,8 +300,7 @@ parse_transform_expr (struct transform_list *tlist, const char *expr,
     {
       if (expr[i] == '\\')
 	{
-	  if (expr[i] == 0)
-	    i += 2;
+	  i += (expr[i] == 0) ? 1 : 2;
 	}
       else if (expr[i] == '[')
 	{
