@@ -336,40 +336,3 @@ mf_priv_setup(struct mf_privs *privs)
 		exit(EX_SOFTWARE);
 	mf_gid_list_free(grp);
 }
-
-
-void
-mf_epriv_setup(struct mf_privs *privs)
-{
-	uid_t uid;
-	gid_t gid;
-	
-	if (privs) {
-		struct passwd *pw;
-		if (!privs->user)
-			return;
-
-		pw = getpwnam(privs->user);
-		if (!pw) {
-			mu_error(_("No such user: %s"), privs->user);
-			exit (EX_CONFIG);
-		}
-		uid = pw->pw_uid;
-		gid = pw->pw_gid;
-	} else {
-		uid = 0;
-		gid = 0;
-	}
-	
-	if (setegid(gid)) {
-		mu_error(_("cannot switch to EGID %lu: %s"),
-			 (unsigned long) gid, mu_strerror(errno));
-		exit(EX_USAGE);
-	}
-	if (seteuid(uid)) {
-		mu_error(_("cannot switch to EUID %lu: %s"),
-			 (unsigned long) uid, mu_strerror(errno));
-		exit(EX_USAGE);
-	}
-}
-
