@@ -123,42 +123,6 @@ config_cb_ignore(void *data, mu_config_value_t *val)
 }
 
 int
-config_cb_lock_retry_count(void *data, mu_config_value_t *val)
-{
-	int rc;
-	char *errmsg;
-	size_t v;
-	
-	if (mu_cfg_assert_value_type(val, MU_CFG_STRING))
-		return 1;
-	rc = mu_str_to_c(val->v.string, mu_c_size, &v, &errmsg);
-	if (rc) {
-		mu_error(_("%s: not a valid number"), val->v.string);
-		free(errmsg);
-		return 1;
-	}
-	mu_locker_set_default_retry_count(v);
-	return 0;
-}
-
-int
-config_cb_lock_retry_timeout(void *data, mu_config_value_t *val)
-{
-	int rc;
-	char *errmsg;
-	time_t v;
-	
-	rc = mu_str_to_c(val->v.string, mu_c_time, &v, &errmsg);
-	if (rc) {
-		mu_error(_("%s: not a valid interval"), val->v.string);
-		free(errmsg);
-		return 1;
-	}
-	mu_locker_set_default_retry_timeout(v);
-	return 0;
-}
-
-int
 stderr_closed_p()
 {
 	int fd = dup(0);
