@@ -53,14 +53,6 @@ static char *resolv_conf_file;
 struct mu_cfg_param callout_cfg_param[] = {
  	{ ".mfd:server",  mu_cfg_section, NULL, 0, NULL, NULL },
 	{ "resolver", mu_cfg_section, NULL, 0, NULL, NULL },
-        { "lock-retry-count", mu_cfg_callback, NULL, 0,
-	  config_cb_lock_retry_count,
-	  N_("Retry acquiring DBM file lock this number of times."),
-          N_("arg: number") },
-	{ "lock-retry-timeout", mu_cfg_callback, NULL, 0,
-	  config_cb_lock_retry_timeout,
-	  N_("Set the time span between the two DBM locking attempts."),
-	  N_("time: interval") },	
 	{ NULL }
 };
 
@@ -105,6 +97,7 @@ main(int argc, char **argv)
 	mu_alloc_die_hook = callout_alloc_die;
 	/* Set default logging */
 	mu_log_facility = DEFAULT_LOG_FACILITY;
+	mu_log_print_severity = 1;
 	mu_stdstream_setup(MU_STDSTREAM_RESET_NONE);
 	mf_srvcfg_log_setup(stderr_closed_p() ? "syslog" : "stderr");
 
