@@ -30,7 +30,18 @@ enum {
 #define DKIM_KEYRECORD_VERSION "DKIM1"
 #define DKIM_SIGNATURE_HEADER "DKIM-Signature"
 #define DKIM_QUERY_METHOD "dns/txt"
-#define DKIM_ALGORITHM "rsa-sha256"
+#define DKIM_KEY_TYPE "rsa"
+
+enum dkim_algorithm {
+	DKIM_ALGO_RSA_SHA1,
+	DKIM_ALGO_RSA_SHA256
+};
+
+#define DKIM_ALGO_DEFAULT DKIM_ALGO_RSA_SHA256
+
+extern char const *dkim_algo_str[];
+int dkim_algo(char const *a);
+
 /* Special value for the signature "l" member */
 #define DKIM_LENGTH_ALL ((size_t)~0)
 
@@ -41,6 +52,7 @@ enum {
  */
 struct dkim_signature {
 	char *a;
+	int algo;
 	uint8_t *b;
 	uint8_t *bh;
 	int canon[2];
@@ -98,6 +110,7 @@ enum {
 	DKIM_EXPL_BAD_BODY	      = _MFL_DKIM_EXPL_BAD_BODY,
 	DKIM_EXPL_BAD_BASE64	      = _MFL_DKIM_EXPL_BAD_BASE64,
 	DKIM_EXPL_BAD_SIG             = _MFL_DKIM_EXPL_BAD_SIG,
+	DKIM_EXPL_BAD_KEY_TYPE        = _MFL_DKIM_EXPL_BAD_KEY_TYPE,
 };
 
 /* Verification error codes */
@@ -107,7 +120,7 @@ enum {
 	DKIM_VERIFY_TEMPFAIL = _MFL_DKIM_VERIFY_TEMPFAIL,
 };
 
-int mfd_dkim_verify(mu_message_t msg, char **sigstr);
+int mfd_dkim_verify(mu_message_t msg, char **sigstr, char const **algo);
 
 extern char const *dkim_explanation_str[];
 extern int dkim_result_trans[];
