@@ -1163,6 +1163,75 @@ code_type_header(NODE *node, struct mu_locus_range const **old_locus)
 	code_immediate(node->v.hdr.name->off, size);
 }
 
+
+/*
+ * Generate code for passing actual number of arguments when calling a
+ * variadic function or function with optional arguments.
+ *
+ * Arguments:
+ *   n    - computed number of arguments
+ *   argv - if argument list ends with $@ - a node_type_argv node describing
+ *          the $@ context, otherwise NULL.
+ *
+ * Calling sequence:
+ *   The function must be called right after the code for passing arguments
+ *   has been generated.  In particular this ensures that code for $@ is
+ *   already generated, so that we can assume that the machine register is
+ *   initialized with shift count (see code_type_argv).
+ *
+ * The code for passing the argument count is as follows:
+ *
+ *  ; Push the count of mandatory and optional arguments
+ *  push n
+ *  ; The code below is generated only if argv != NULL
+ *  ; A. Push the actual number of arguments to the calling function
+ *  memstk 0(3)
+ *  deref
+ *  add
+ *  ; B. Push the number of formal arguments to the calling function
+ *  push pargs
+ *  ; subtract: A - B.  This leaves on stack number of variadic arguments
+ *  ; passed to the calling function.
+ *  sub
+ *  ; Register keeps shift count - N from $@(N)
+ *  ; Push it on stack
+ *  pushreg
+ *  ; Subtract this value from argument count
+ *  sub
+ */
+static void
+code_hidden_argcount(long n, NODE *argv)
+{
+	code_op(opcode_push);
+	code_immediate(n, long);
+	if (argv) {
+		/*
+		 * Add number of arguments passed to the calling
+		 * function.
+		 */
+		code_op(opcode_memstk);
+		code_immediate(0, size);
+		code_immediate(3, long);
+		code_op(opcode_deref);
+		code_op(opcode_add);
+		/*
+		 * Minus number of non-variable arguments.
+		 */
+		if (argv->v.argv.nargs) {
+			code_op(opcode_push);
+			code_immediate(argv->v.argv.nargs, long);
+			code_op(opcode_sub);
+		}
+
+		/*
+		 * Minus shift count, if any.
+		 */
+		if (argv->v.argv.node) {
+			code_op(opcode_pushreg);
+			code_op(opcode_sub);
+		}
+	}
+}
 
 /* type builtin */
 
@@ -1214,6 +1283,51 @@ optimize_type_builtin(NODE *node)
 	}
 }
 
+/*
+ * Produce code for passing argument list P, whose first argument represents
+ * a shifted arglist $@(N).  The caller should ensure that this is the case
+ * before calling this function.
+ *
+ * Return value: number of arguments (words) pushed on stack, not
+ * counting the ones generated by $@(N) instruction.
+ *
+ * This function is called by code_type_call and code_type_builtin.
+ *
+ * Note: The algorithm below is suboptimal, but it seems to be the only way
+ * given scarcity of registers.  I need to implement at least one more
+ * register in order to do that efficiently.  So, for now:
+ */
+static size_t
+code_args_shifted(NODE *p)
+{
+	size_t i;
+	
+	/*
+	 * First, process the $@(N) node.  This will, among others, generate
+	 * code for leaving shift count in the register.
+	 */
+	code_node(p);
+	i = 0;
+	if ((p = p->next) != NULL) {
+		/*
+		 * Code the rest of arguments.  Eventual function calls
+		 * appearing in these will clobber the reg, so push it on
+		 * on stack first and keep exchanging with tos after each
+		 * subsequent argument is coded, so that shift count always
+		 * remains on top.
+		 */
+		code_op(opcode_pushreg);
+		do {
+			code_node(p);
+			code_op(opcode_xchg);
+			i++;
+		} while ((p = p->next) != NULL);
+		/* Finally, pop it back to the register. */
+		code_op(opcode_popreg);
+	}
+	return i;
+}
+
 void
 code_type_builtin(NODE *node, struct mu_locus_range const **old_locus)
 {
@@ -1221,14 +1335,24 @@ code_type_builtin(NODE *node, struct mu_locus_range const **old_locus)
 	int i;
 	const struct builtin *bp = node->v.builtin.builtin;
 	struct literal *s;
-
+	NODE *argv = (node->v.builtin.args &&
+		      node->v.builtin.args->type == node_type_argv) ?
+		      node->v.builtin.args : NULL;
+	
 	/* Pass arguments */
-	for (p = node->v.builtin.args, i = 0; p; p = p->next, i++)
-		code_node(p);
+	if (argv && argv->v.argv.node) {
+		/* Shift count given: $@(N).  */
+		i = code_args_shifted(argv);
+	} else {
+		for (p = node->v.builtin.args, i = 0; p; p = p->next, i++) {
+			code_node(p);
+		}
+		if (argv) i--;
+	}
+		    
 	if (bp->optcount || (bp->flags & MFD_BUILTIN_VARIADIC)) {
 		/* Pass the number of actual arguments in a hidden arg */
-		code_op(opcode_push);
-		code_immediate(i, size);
+		code_hidden_argcount(i, argv);
 	}
 	
 	mark_locus(node, old_locus);
@@ -1236,6 +1360,11 @@ code_type_builtin(NODE *node, struct mu_locus_range const **old_locus)
 	s = literal_lookup(node->v.builtin.builtin->name);
 	code_immediate(s->off, size);
 	code_immediate(node->v.builtin.builtin->handler, ptr);/*FIXME*/
+
+	if (node->value_ignored) {
+		code_op(opcode_adjust);
+		code_immediate(1, ulong);
+	}
 }
 
 
@@ -1443,7 +1572,87 @@ optimize_type_argx(NODE *node)
 		node->v.arg.number = argx->v.number + nargs;
 	}
 }
+
+/* type argv */
+void
+print_type_argv(NODE *node, int level)
+{
+	print_level(level);
+	printf("ARGV\n");
+	if (node->v.argv.node)
+		print_node(node->v.argv.node, level + 1);
+}
 
+void
+mark_type_argv(NODE *node)
+{
+       mark(node->v.argv.node);
+}
+
+void
+optimize_type_argv(NODE *node)
+{
+       optimize(node->v.argv.node);
+}
+
+void
+code_type_argv(NODE *node, struct mu_locus_range const **old_locus)
+{
+	mark_locus(node, old_locus);
+
+	/*
+	 * If shift count is given, generate code for it.  Duplicate it
+	 * on stack and leave a copy of it in register.  This will later
+	 * be used in code generated by code_hidden_argcount.
+	 */
+	if (node->v.argv.node) {
+		code_node(node->v.argv.node);
+		code_op(opcode_dup);
+		code_op(opcode_dup);
+		code_op(opcode_popreg);
+	}
+
+	/*
+	 * Second argument: number of stack words to push
+	 */
+	code_op(opcode_memstk);
+	code_immediate(0, size);
+	code_immediate(3, long);
+	code_op(opcode_deref);
+	/* Subtract number of formal arguments. */
+	if (node->v.argv.nargs) {
+		code_op(opcode_push);
+		code_immediate(node->v.argv.nargs, long);
+		code_op(opcode_sub);
+	}
+	/*
+	 * If shift count is given, tos+1 is the count computed so far,
+	 * tos+2 and tos+3 keep the shift count.
+	 * Swap two topmost entries and subtract.  Swap the result with
+	 * the shift count so it becomes the topmost entry.
+	 */
+	if (node->v.argv.node) {
+		code_op(opcode_xchg);
+		code_op(opcode_sub);
+		code_op(opcode_xchg);
+	}
+
+        /* First argument: parameter offset */
+	code_op(opcode_push);
+	code_immediate(node->v.argv.nargs + 4, long);
+
+	/*
+	 * If shift count is given, add it to the offset.
+	 */
+	if (node->v.argv.node)
+		code_op(opcode_add);
+
+	/*
+	 * Finally, push the selected stack entries on stack again.
+	 */
+	code_op(opcode_dupn);
+	code_immediate(0, size);
+}
 
 /* type vaptr */
 void
@@ -1780,10 +1989,11 @@ code_type_try(NODE *node, struct mu_locus_range const **old_locus)
 	struct catch_node *catch = &node->v.try.catch->v.catch;
 
 	ctr = jump_pc;
-	
+
 	code_op(opcode_saveex);
 	code_exmask(catch->exmask);
 	
+	mark_locus(node->v.try.catch, old_locus);
 	code_op(opcode_catch);
 	pos1 = code_immediate(NULL, ptr);
 	code_exmask(catch->exmask);
@@ -1975,7 +2185,7 @@ optimize_type_call(NODE *node)
 	for (p = node->v.call.args; p; p = p->next) 
 		optimize(p);
 }
-		
+
 void
 code_type_call(NODE *node, struct mu_locus_range const **old_locus)
 {
@@ -1986,10 +2196,16 @@ code_type_call(NODE *node, struct mu_locus_range const **old_locus)
 
 	if (func->optcount || func->varargs) {
 		int j;
+		NODE *argv = (node->v.call.args &&
+			      node->v.call.args->type == node_type_argv) ?
+			      node->v.call.args : NULL;
 		
 		/* Count actual arguments */
 		for (p = node->v.call.args, i = 0; p; p = p->next, i++)
 			;
+
+		if (argv) i--;
+		
 		/* Provide placeholders for the missing ones.  This is
 		   necessary in order to make assignments to optional
 		   arguments possible. */
@@ -1999,29 +2215,44 @@ code_type_call(NODE *node, struct mu_locus_range const **old_locus)
 		}
 
 		/* Pass actual arguments */
-		for (p = node->v.call.args; p; p = p->next) 
-			code_node(p);
-		
-		/* Pass the number of actual arguments in a hidden arg */
-		code_op(opcode_push);
-		code_immediate(i, long);
-	} else
+		if (argv && argv->v.argv.node) {
+			/* Shift count given: $@(N).  */
+			code_args_shifted(argv);
+		} else {
+			/* No shifts involved: everything is straightforward */
+			for (p = node->v.call.args; p; p = p->next) {
+				code_node(p);
+			}
+		}
+
+		code_hidden_argcount(i, argv);
+	} else {
+		/*
+		 * Function takes fixed number of parameters: everything is
+		 * straightforward as well.
+		 */
 		for (p = node->v.call.args, i = 0; p; p = p->next, i++) 
 			code_node(p);
+	}
 	
 	mark_locus(node, old_locus);
-	if (func->entry == 0) {
-		parse_error("INTERNAL ERROR at %s:%d: unresolved function %s",
-			    __FILE__, __LINE__, func->sym.name);
-		abort();
-	}
 	code_op(opcode_funcall);
 	s = literal_lookup(func->sym.name);
 	code_immediate(s->off, size);
-	code_immediate(func->entry, size);
-	code_op(opcode_adjust);
-	code_immediate((func->parmcount + (func->optcount ? 1 : 0)), ulong);
-	if (func->rettype != dtype_unspecified)
+	if (func->entry == 0) {
+		func->refstart = code_immediate(func->refstart, size);
+	} else {
+		code_immediate(func->entry, size);
+	}
+
+	if (func->varargs) {
+		code_op(opcode_adjustx);
+		code_immediate(func->parmcount, ulong);
+	} else {
+		code_op(opcode_adjust);
+		code_immediate((func->parmcount + (func->optcount ? 1 : 0)), ulong);
+	}
+	if (func->rettype != dtype_unspecified && !node->value_ignored)
 		code_op(opcode_pushreg);
 }
 
@@ -2598,20 +2829,32 @@ optimize_type_cast(NODE *node)
 
 	optimize(node->v.cast.node);
 	p = node->v.cast.node;
-	if (p->type == node_type_string) {
-		node->type = node_type_string;
+	if (node->v.cast.data_type == node_type(p)) {
+		node->type = p->type;
+		node->locus = p->locus;
 		node->v = p->v;
 		free_node(p);
-	} else if (p->type == node_type_number) {
+	} else if (node->v.cast.data_type == dtype_string &&
+		   p->type == node_type_number) {
 		char buf[NUMERIC_BUFSIZE_BOUND];
 		snprintf(buf, sizeof buf, "%ld", p->v.number);
 		node->v.literal = string_alloc(buf, strlen(buf));
 		node->type = node_type_string;
-	} else if (node->v.cast.data_type == dtype_number
-		   && node_type(p) == dtype_number) {
-		node->type = p->type;
+	} else if (node->v.cast.data_type == dtype_number &&
+		   p->type == node_type_string) {
+		long val;
+		char *endp;
+		errno = 0;
+		val  = strtol(p->v.literal->text, &endp, 0);
+		if (errno || *endp) {
+			parse_error_locus(&node->locus,
+					  _("cannot convert literal to number"));
+			return;
+		}
+
+		node->type = node_type_number;
 		node->locus = p->locus;
-		node->v = p->v;
+		node->v.number = val;
 		free_node(p);
 	}
 }
@@ -2637,6 +2880,8 @@ code_cast(struct mu_locus_range const *locus, data_type_t fromtype, data_type_t
 		code_op(opcode_ntos);
 		break;
 
+	case dtype_any:
+		break;
 	}
 	return 0;
 }
@@ -2674,8 +2919,19 @@ code_type_funcdecl(NODE *node, struct mu_locus_range const **old_locus)
 	   handle recursive functions */
 	func->entry = code_get_counter();
 			
-	codegen(&pc, node->v.funcdecl.tree, func->exmask, 0,
+	codegen(&pc, node->v.funcdecl.tree, func->exmask,
 		node->v.funcdecl.auto_count);
+
+	/*
+	 * Fix-up unresolved references.
+	 */
+	pc = func->refstart;
+	while (pc) {
+		prog_counter_t next = (prog_counter_t)mf_cell_c_value(code_peek(pc), size);
+		code_put(pc, func->entry, size);
+		pc = next;
+	}
+	func->refstart = 0;
 }
 	
 
@@ -2684,33 +2940,21 @@ code_type_funcdecl(NODE *node, struct mu_locus_range const **old_locus)
 void
 mark_type_progdecl(NODE *node)
 {
-	mark(node->v.progdecl.tree);
+	mark(node->v.progdecl.tree.head);
 }
 		
 void
 optimize_type_progdecl(NODE *node)
 {
-	optimize(node->v.progdecl.tree);
+	optimize(node->v.progdecl.tree.head);
 }
 
 void
 code_type_progdecl(NODE *node, struct mu_locus_range const **old_locus)
 {
-	enum smtp_state tag;
-
-	tag = node->v.progdecl.tag;
-
-	if (root_node[tag]) {
-		parse_warning_locus(&node->locus,
-				    _("redefinition of handler `%s'"),
-				    state_to_string(tag));
-		parse_warning_locus(&root_node[tag]->locus,
-				    _("this is the location of the previous definition"));
-	}
-				  
-	root_node[tag] = node->v.progdecl.tree;
+	enum smtp_state tag = node->v.progdecl.tag;
 	if (codegen(&entry_point[tag],
-		    node->v.progdecl.tree, 0, 1,
+		    node->v.progdecl.tree.head, 0,
 		    node->v.progdecl.auto_count) == 0)
 		milter_enable_state(tag);
 }
