@@ -37,17 +37,17 @@
 #endif
 
 /* Exception codes. */
-typedef enum mf_exception_code {
-#include <mflib/status.ex>
-	mf_exception_count
-} mf_exception;
+#include <mflib/exceptions.h>
 
 /* SMTP states */
 enum smtp_state {
 	smtp_state_none,
+	smtp_state_startup,
+	smtp_state_shutdown,
 	smtp_state_begin,
-	smtp_state_first=smtp_state_begin,
+	smtp_state_end,
 	smtp_state_connect,
+	smtp_state_first=smtp_state_connect,
 	smtp_state_helo,
 	smtp_state_envfrom,
 	smtp_state_envrcpt,
@@ -56,9 +56,8 @@ enum smtp_state {
 	smtp_state_eoh,
 	smtp_state_body,
 	smtp_state_eom,
+	smtp_state_last=smtp_state_eom,
 
-	smtp_state_end,
-	
 	smtp_state_count
 };
 
@@ -90,7 +89,17 @@ extern mu_debug_handle_t mfd_debug_handle;
 /* FIXME */
 void enable_prog_trace(const char *modlist);
 void disable_prog_trace(const char *modlist);
+
+struct pathbuf {
+	char *base;   /* Constructed path */
+	int pos;      /* Insert position in base */
+	int alloc;    /* 1 if base was allocated */
+};
 
+#define PATHBUF_INITIALIZER(p) \
+	{ .base = p, .pos = 0, .alloc = 0 }
+
+void pathbuf_add(struct pathbuf *pb, char const *dir);
 
 /* Global variables declarations */
 #define MAILFROMD_DAEMON        0
@@ -99,7 +108,6 @@ void disable_prog_trace(const char *modlist);
 #define MAILFROMD_RUN           3
 
 extern char *script_file;
-extern char *ext_pp;
 extern char *ext_pp_options;
 extern int ext_pp_options_given;
 extern int location_column_option;
@@ -114,6 +122,10 @@ extern int script_dump_xref;
 extern int stack_trace_option;
 extern const char *program_version;
 extern char *callout_server_url;
+extern struct pathbuf include_path;
+extern struct pathbuf module_path;
+extern char *mfmod_path;
+extern char const *default_suffixes[];
 
 extern size_t stack_size;
 extern size_t stack_max_size;
@@ -133,12 +145,18 @@ typedef enum {
 	dtype_unspecified,
 	dtype_string,
 	dtype_number,
-	dtype_pointer
+	dtype_pointer,
+	dtype_any,
 } data_type_t;
 
 /* Parse tree node */
 typedef struct node NODE;
 
+struct stmtlist {
+	NODE *head;
+	NODE *tail;
+};
+
 typedef struct mu_locus_range mf_yyltype_t;
 
 struct if_node {            /* if statement */
@@ -315,7 +333,7 @@ struct positional_arg {
 struct progdecl {
 	enum smtp_state tag;
 	size_t auto_count;
-	NODE *tree;
+	struct stmtlist tree;
 };
 
 struct funcdecl {
@@ -365,13 +383,21 @@ struct sed_node {
 	NODE *arg;
 };
 
+struct argv_node {
+	size_t nargs;   /* Number of fixed arguments */
+	NODE *node;
+};
+	
 #include "node-type.h"
 
 /* Parse tree node */
 struct node {
 	NODE *next;                   /* Link to the next node */
 	enum node_type type;          /* Node type */
-	struct mu_locus_range locus;           /* Corresponding input location */
+	struct mu_locus_range locus;  /* Corresponding input location */
+	int value_ignored;            /* Is the node value ignored; curently
+					 only for node_type_funcall and
+					 node_type_builtin */
 	union {
 		struct literal *literal;
 		long number;
@@ -401,14 +427,10 @@ struct node {
 		struct argx_node argx;
 		struct sedcomp_node sedcomp;
 		struct sed_node sed;
+		struct argv_node argv;
 	} v;
 };
 
-struct stmtlist {
-	NODE *head;
-	NODE *tail;
-};
-
 struct exmask {
 	struct exmask *next;
 	struct bitmask bm;
@@ -482,8 +504,10 @@ struct function {      /* User-defined function */
 	size_t optcount;         /* Number of optional parameters */
 	int varargs;             /* 1 if function takes variable number of
 				    arguments */
+	data_type_t vartype;     /* Type to promote varargs to */
 	data_type_t *parmtype;   /* Parameter types */
 	data_type_t rettype;     /* Return type */
+	prog_counter_t refstart; /* Start of the unresolved entrypoint chain */ 
 };	
 
 typedef enum {
@@ -512,7 +536,7 @@ struct literal {
 	SYMENT_STRUCT(text);
 	unsigned flags;            /* SYM_* flags */
 	size_t off;                /* Offset in DS */
-	struct sym_regex *regex;   /* Any correspondig regexes */
+	struct sym_regex *regex;   /* Any corresponding regexes */
 };
 
 struct constant {
@@ -576,6 +600,7 @@ int variable_or_constant_lookup(const char *name, void **dptr);
 
 struct function *function_install(const char *name,
 				  size_t parmcnt, size_t optcnt, int varargs,
+				  data_type_t vartype,
 				  data_type_t *parmtypes,
 				  data_type_t rettype,
 				  const struct mu_locus_range *locus);
@@ -688,16 +713,29 @@ int module_symtab_enumerate(enum module_namespace ns, symtab_enumerator_t fun,
 
 
 /* Preprocessor functions */
+struct preprocessor {
+	int enabled;
+	char *command;
+	int pass_includes;
+	int pass_defines;
+	char *setup_file;
+};
+
 struct input_file_ident {
 	ino_t i_node;
 	dev_t device;
 };
 
-void include_path_setup(void);
+extern struct preprocessor preprocessor;
 
+void preprocessor_cfg_init(void);
 FILE *pp_extrn_start(const char *name, pid_t *ppid);
 void pp_extrn_shutdown(FILE *file, pid_t pid);
 int preprocess_input(void);
+char const *string_preprocessor(void);
+void preprocessor_finalize(void);
+
+char *path_find_file(const char *path, const char *name, char const **suffixes);
 
 void require_module(const char *modname, struct import_rule *import_rules);
 
@@ -705,7 +743,6 @@ int parse_include(const char *text, int once);
 int parse_require(const char *text, mf_yyltype_t *loc);
 int parse_line(char *text, struct mu_locus_point *pt);
 int parse_line_cpp(char *text, struct mu_locus_point *pt);
-void alloc_ext_pp(void);
 int source_lookup(struct input_file_ident *idptr);
 
 #define LEX_NONE   0
@@ -722,7 +759,6 @@ int yyparse();
 int yylex();
 int yyerror(char const *s);
 void free_parser_data(void);
-void add_include_dir(const char *dir);
 int parse_program(char *name, int ydebug);
 void parse_pragma(const char *text);
 struct mu_locus_range const *get_locus(void);
@@ -825,7 +861,6 @@ void env_push_string(eval_environ_t env, char *arg);
 void env_push_number(eval_environ_t env, long arg);
 void env_push_pointer(eval_environ_t env, void *arg);
 void env_make_frame(eval_environ_t env);
-void env_make_frame0(eval_environ_t env);
 void env_leave_frame(eval_environ_t env, int nargs);
 int env_set_variable(eval_environ_t env, char *ident, char *value);
 int eval_environment(eval_environ_t env, prog_counter_t pc);
@@ -880,6 +915,7 @@ void milter_setlogmask(int mask);
 void milter_settimeout(time_t t);
 
 int xeval(eval_environ_t env, enum smtp_state tag);
+int xeval_begin(eval_environ_t env);
 
 int relayed_domain_p(char *name);
 size_t mem_search(const char *str, int c, size_t size);
