@@ -320,14 +320,6 @@ struct mu_cfg_param mfdbtool_cfg_param[] = {
 	     "and Debugging\", GNU Mailutils manual, section 3.3 \"Debugging\",\n"
 	     "or visit <http://mailutils.org/wiki/Debug_level>."),
 	  N_("spec: list") },
-        { "lock-retry-count", mu_cfg_callback, NULL, 0,
-	  config_cb_lock_retry_count,
-	  N_("Retry acquiring DBM file lock this number of times."),
-	  N_("arg: number")},
-	{ "lock-retry-timeout", mu_cfg_callback, NULL, 0,
-	  config_cb_lock_retry_timeout,
-	  N_("Set the time span between the two DBM locking attempts."),
-	  N_("time: interval") },
 	
 	{ NULL }
 };
