@@ -22,25 +22,106 @@
 #include "mailfromd.h"
 #include "prog.h"
 
-extern int yy_flex_debug;
+void
+pathbuf_add(struct pathbuf *pb, char const *dir)
+{
+	size_t dir_len = strlen(dir);
+	if (pb->alloc) {
+		size_t init_len = strlen(pb->base);
+		int delim = pb->base[pb->pos] != 0;
+		pb->base = mu_realloc(pb->base, init_len + 1 + dir_len + 1);
+		if (delim) {
+			memmove(pb->base + pb->pos + dir_len + 1, pb->base + pb->pos,
+				init_len - pb->pos + 1);
+			memcpy(pb->base + pb->pos, dir, dir_len);
+			pb->pos += dir_len;
+			pb->base[pb->pos++] = ':';
+		} else {
+			pb->base[pb->pos++] = ':';
+			strcpy(pb->base + pb->pos, dir);
+			pb->pos += dir_len;
+		}
+	} else if (pb->base) {
+		size_t init_len = strlen(pb->base);
+		char *p = mu_alloc(init_len + 1 + dir_len + 1);
+		memcpy(p, dir, dir_len);
+		pb->pos += dir_len;
+		p[pb->pos++] = ':';
+		strcpy(p + pb->pos, pb->base);
+		pb->base = p;
+		pb->alloc = 1;
+	} else {
+		pb->base = strdup(dir);
+		pb->pos = strlen(pb->base);
+		pb->alloc = 1;
+	}
+}
+
+char const *default_suffixes[] = { ".mfl", ".mf", NULL };
 
-static mu_list_t include_path;
-static mu_list_t std_include_path;
+struct pathbuf include_path = PATHBUF_INITIALIZER(
+	DEFAULT_MODDIR "/include:"
+	DEFAULT_VERSION_MODDIR "/include"
+);
 
-struct file_data {
+struct pathbuf module_path = PATHBUF_INITIALIZER(
+        DEFAULT_MODDIR ":" DEFAULT_VERSION_MODDIR
+);
+
+static void
+path_enumerate(char const *path, int (*func)(const char *, void *), void *data)
+{
+	char *dirbuf = NULL;
+	size_t dirsize = 0;
+
+	if (!path)
+		return;
+	for (;;) {
+		size_t len = strcspn(path, ":");
+		len++;
+		if (len > dirsize) {
+			dirsize = len;
+			dirbuf = mu_realloc(dirbuf, dirsize);
+		}
+		len--;
+		memcpy(dirbuf, path, len);
+		dirbuf[len] = 0;
+
+		if (func(dirbuf, data))
+			break;
+
+		path += len;
+		if (*path)
+			path++;
+		else
+			break;
+	}
+	free(dirbuf);
+}
+
+struct scan_buffer {
 	const char *name;
 	size_t namelen;
+	char const **suffixes;
 	char *buf;
 	size_t buflen;
 	int found;
 };
 
+#define SCAN_BUFFER_INITIALIZER(n, s)			\
+	{						\
+	        .name = n,				\
+		.namelen = strlen(n),			\
+		.suffixes = s,				\
+		.buf = NULL,				\
+		.buflen = 0,				\
+		.found = 0				\
+	}
+
 static int
-find_include_file(void *item, void *data)
+test_file(char const *dir, struct scan_buffer *dptr, char const *suf)
 {
-	char *dir = item;
-	struct file_data *dptr = data;
-	size_t size = strlen(dir) + 1 + dptr->namelen + 1;
+	size_t size = strlen(dir) + dptr->namelen + (suf ? strlen(suf) : 0) + 2;
 	if (size > dptr->buflen) {
 		dptr->buflen = size;
 		dptr->buf = mu_realloc(dptr->buf, dptr->buflen);
@@ -48,39 +129,64 @@ find_include_file(void *item, void *data)
 	strcpy(dptr->buf, dir);
 	strcat(dptr->buf, "/");
 	strcat(dptr->buf, dptr->name);
-	return dptr->found = access(dptr->buf, F_OK) == 0;
+	if (suf)
+		strcat(dptr->buf, suf);
+	return dptr->found = (access(dptr->buf, F_OK) == 0);
 }
 
-void
-include_path_setup()
+static int
+scan_dir_fun(char const *dir, void *data)
 {
-	mu_list_create(&include_path);
-	mu_list_create(&std_include_path);
-	mu_list_append(std_include_path, DEFAULT_VERSION_INCLUDE_DIR);
-	mu_list_append(std_include_path, DEFAULT_INCLUDE_DIR);
-	mu_list_append(std_include_path, "/usr/share/mailfromd/include");
-	mu_list_append(std_include_path, "/usr/local/share/mailfromd/include");
-}
+	struct scan_buffer *dptr = data;
 
-void
-add_include_dir(const char *dir)
-{
-	int rc;
-	
-	if (!include_path) {
-		rc = mu_list_create(&include_path);
-		if (rc) {
-			mu_error(_("cannot create include path list: %s"),
-				 mu_strerror(rc));
-			return;
+	if (dptr->suffixes) {
+		int i;
+
+		for (i = 0; dptr->suffixes[i]; i++) {
+			if (test_file(dir, dptr, dptr->suffixes[i]))
+				break;
 		}
-	}
-	rc = mu_list_append(include_path, mu_strdup(dir));
-	if (rc) 
-		mu_error(_("cannot append to the include path list: %s"),
-			 mu_strerror(rc));
+	} else
+		test_file(dir, dptr, NULL);
+	return dptr->found;
 }
 
+char *
+path_find_file(const char *path, const char *name, char const **suffixes)
+{
+	struct scan_buffer sbuf = SCAN_BUFFER_INITIALIZER(name, suffixes);
+	path_enumerate(path, scan_dir_fun, &sbuf);
+	if (sbuf.found)
+		return sbuf.buf;
+	free(sbuf.buf);
+	return NULL;
+}
+
+static int
+try_file(const char *name, const char **suffixes,
+	 struct pathbuf *pb,
+	 int allow_cwd, int err_not_found, char **newp)
+{
+	struct scan_buffer sbuf = SCAN_BUFFER_INITIALIZER(name, suffixes);
+
+	if (allow_cwd) {
+		if ((*newp = path_find_file(".", name, suffixes)) != NULL)
+			return 1;
+	}
+
+	path_enumerate(pb->base, scan_dir_fun, &sbuf);
+
+	if (!sbuf.found && err_not_found) {
+		parse_error(_("%s: no such file or directory"), name);
+		*newp = NULL;
+	}
+
+	if (sbuf.found)
+		*newp = sbuf.buf;
+	else
+		free(sbuf.buf);
+	return sbuf.found;
+}
 
 #define ID_EQ(a,b)						\
 	((a).i_node == (b).i_node && (a).device == (a).device)
@@ -119,38 +225,6 @@ source_lookup(struct input_file_ident *idptr)
 	return rc == 0;
 }
 
-static int
-try_file(const char *name, int allow_cwd, int err_not_found, char **newp)
-{
-	static char *cwd = ".";
-	struct file_data fd;
-			
-	fd.name = name;
-	fd.namelen = strlen(name);
-	fd.buf = NULL;
-	fd.buflen = 0;
-	fd.found = 0;
-	
-	if (allow_cwd) {
-		mu_list_prepend(include_path, cwd);
-		mu_list_foreach(include_path, find_include_file, &fd);
-		mu_list_remove(include_path, cwd);
-	} else
-		mu_list_foreach(include_path, find_include_file, &fd);
-			
-	if (!fd.found) {
-		mu_list_foreach(std_include_path, find_include_file, &fd);
-					
-		if (!fd.found && err_not_found) {
-			parse_error(_("%s: no such file or directory"), name);
-			*newp = NULL;
-		} 
-	}
-	if (fd.found)
-		*newp = fd.buf;
-	return fd.found;
-}
-
 int
 begin_module(const char *modname, const char *filename,
 	     struct import_rule *import_rules)
@@ -180,10 +254,10 @@ parse_include(const char *text, int once)
 
 	if (mu_wordsplit(text, &ws, MU_WRDSF_DEFFLAGS)) {
 		parse_error(_("cannot parse include line: %s"),
-			    mu_wordsplit_strerror (&ws));
+			    mu_wordsplit_strerror(&ws));
 		return errno;
 	} else if (ws.ws_wordc != 2)
-		mu_wordsplit_free (&ws);
+		mu_wordsplit_free(&ws);
 	else {
 		size_t len;
 		int allow_cwd;
@@ -198,34 +272,32 @@ parse_include(const char *text, int once)
 		} else 
 			allow_cwd = 1;
 
-		if (p[0] != '/' && try_file(p, allow_cwd, 1, &tmp))
+		if (p[0] != '/' && try_file(p, NULL, &include_path, allow_cwd, 1, &tmp))
 			p = tmp;
 	}
 
 	if (p)
 		rc = lex_new_source(p, once ? LEX_ONCE : LEX_NONE);
 	free(tmp);
-	mu_wordsplit_free (&ws);
+	mu_wordsplit_free(&ws);
 	return rc;
 }
 
-#define DEFAULT_SUFFIX ".mf"
-
 void
 require_module(const char *modname, struct import_rule *import_rules)
 {
-	size_t len;
-	char *fname, *pathname;
+	char *pathname;
 	
-	len = strlen(modname);
-	
-	fname = mu_alloc(len + sizeof DEFAULT_SUFFIX);
-	strcpy(fname, modname);
-	strcat(fname, DEFAULT_SUFFIX);
-			
-	if (try_file(fname, 0, 1, &pathname))
-		begin_module(modname, pathname, import_rules);
-	free(fname);
+	if (try_file(modname, default_suffixes, &module_path, 0, 0, &pathname) == 0) {
+		if (try_file(modname, default_suffixes, &include_path, 0, 1, &pathname) == 0) {
+			return;
+		} else {
+			parse_warning(_("module %s found in deprecated location: include path (%s)"), modname, pathname);
+			parse_warning(_("please move it somewhere to your module path (%s)"),
+				      module_path.base);
+		}
+	}
+	begin_module(modname, pathname, import_rules);
 }
 
 static int
@@ -299,10 +371,51 @@ parse_line_cpp(char *text, struct mu_locus_point *ploc)
 	mu_wordsplit_free(&ws);
 	return rc;
 }
+
+struct preprocessor preprocessor = {
+	.enabled = 1,
+	.command = DEF_EXT_PP,
+	.pass_includes = 0,
+	.pass_defines = 1,
+	.setup_file = "pp-setup"
+};
 
-static void
-stderr_redirector(char *ppcmd, char **argv)
+static struct mu_cfg_param preprocessor_section_param[] = {
+	{ "enable", mu_c_bool, &preprocessor.enabled, 0, NULL,
+	  N_("Enable preprocessor (default).") },
+	{ "command", mu_c_string, &preprocessor.command, 0, NULL,
+	  N_("Preprocessor command line stub.") },
+	{ "pass-includes", mu_c_bool, &preprocessor.pass_includes, 0, NULL,
+	  N_("Pass include search path to the preprocessor via -I options (default: false)") },
+	{ "pass-defines", mu_c_bool, &preprocessor.pass_defines, 0, NULL,
+	  N_("Pass feature definition and defines from the command line to the preprocessor via -D options (default: true)") },
+	{ "setup-file", mu_c_string, &preprocessor.setup_file, 0, NULL,
+	  N_("Name of the preprocessor setup file.  Looked up in include search path, unless absolute.") },
+	{ NULL }
+};
+
+void
+preprocessor_cfg_init(void)
 {
+	struct mu_cfg_section *section;
+	if (mu_create_canned_section("preprocessor", &section) == 0) {
+		section->label = NULL;
+		section->parser = NULL;
+		section->docstring = N_("MFL preprocessor configuration.");
+		mu_cfg_section_add_params(section, preprocessor_section_param);
+	}
+	if (preprocessor.command == NULL)
+		preprocessor.enabled = 0;
+}
+
+static void
+stderr_redirector(char *ppcmd)
+{
+	char *xargv[] = { getenv("SHELL"), "-c", ppcmd, NULL };
+
+	if (!xargv[0])
+		xargv[0] = "/bin/sh";
+
 	if (!logger_flags(LOGF_STDERR)) {
 		int p[2];
 		char buf[1024];
@@ -324,7 +437,7 @@ stderr_redirector(char *ppcmd, char **argv)
 			}
 			close(p[0]);
 			
-			execvp(argv[0], argv);
+			execvp(xargv[0], xargv);
 			exit(127);
 			
 		case -1:
@@ -344,7 +457,7 @@ stderr_redirector(char *ppcmd, char **argv)
 			exit(0);
 		}
 	} else {
-		execvp(argv[0], argv);
+		execvp(xargv[0], xargv);
 		mf_server_log_setup();
 		mu_error("Cannot run `%s': %s",
 			 ppcmd, mu_strerror(errno));
@@ -359,24 +472,17 @@ pp_extrn_start(const char *name, pid_t *ppid)
 	pid_t pid;
 	size_t size;
 	char *ppcmd;
-	struct mu_wordsplit ws;
 	FILE *fp = NULL;
 	
-	size = strlen(ext_pp) + strlen(name) + 2;
+	size = strlen(preprocessor.command) + strlen(name) + 2;
 	ppcmd = mu_alloc(size);
-	strcpy(ppcmd, ext_pp);
+	strcpy(ppcmd, preprocessor.command);
 	strcat(ppcmd, " ");
 	strcat(ppcmd, name);
 
 	mu_debug(MF_SOURCE_PP, MU_DEBUG_TRACE1,
 		 ("Running preprocessor: `%s'", ppcmd));
 
-	if (mu_wordsplit(ppcmd, &ws, MU_WRDSF_DEFFLAGS)) {
-		parse_error(_("cannot parse line: %s"),
-			    mu_wordsplit_strerror (&ws));
-		exit(EX_UNAVAILABLE);
-	}
-	
 	if (pipe(pout)) {
 		mu_diag_funcall(MU_DIAG_ERROR, "pipe", "pout", errno);
 		exit(EX_UNAVAILABLE);
@@ -389,9 +495,9 @@ pp_extrn_start(const char *name, pid_t *ppid)
 			dup2(pout[1], 1);
 		}
 
-                /* Close unneded descripitors */
+                /* Close unneeded descriptors */
 		close_fds_above(2);
-		stderr_redirector(ppcmd, ws.ws_wordv);
+		stderr_redirector(ppcmd);
 		break;
 
 	case -1:
@@ -405,7 +511,6 @@ pp_extrn_start(const char *name, pid_t *ppid)
 		fp = fdopen(pout[0], "r");
 		break;
 	}
-	mu_wordsplit_free(&ws);
 	free(ppcmd);
 	*ppid = pid;
 	return fp;
@@ -434,43 +539,42 @@ pp_extrn_shutdown(FILE *file, pid_t pid)
 }
 
 static int
-_count_include_size(void *item, void *data)
+_count_include_size(const char *dir, void *data)
 {
 	size_t *psize = data;
-	*psize += 3 + strlen((char*)item);
+	*psize += 3 + strlen(dir);
 	return 0;
 }
 
 static int
-_append_includes(void *item, void *data)
+_append_includes(const char *dir, void *data)
 {
 	char *str = data;
 	strcat(str, " -I");
-	strcat(str, (char*) item);
+	strcat(str, dir);
 	return 0;
 }
 
-static char *setup_file;
-
 void
-alloc_ext_pp()
+preprocessor_finalize(void)
 {
 	size_t size;
+	char *setup_file;
+	char *ext_pp;
 	
-	if (ext_pp_options_given && !ext_pp) {
-		if (!DEF_EXT_PP) {
-			mu_error(_("default preprocessor is not set; "
-				   "use --preprocessor option"));
+	if (!preprocessor.enabled || !preprocessor.command) {
+		if (ext_pp_options_given) {
+			mu_error(_("preprocessor options given, "
+				   "but the use of preprocessor is disabled; "
+				   "use the --preprocessor option"));
 			exit(EX_USAGE);
 		}
-		ext_pp = DEF_EXT_PP;
-	}		
-
-	if (!ext_pp)
 		return;
-	ext_pp = mu_strdup(ext_pp);
+	}
 	
-	if (ext_pp_options) {
+	ext_pp = mu_strdup(preprocessor.command);
+	
+	if (ext_pp_options && preprocessor.pass_defines) {
 		size_t len;
 		char *p;
 	
@@ -481,23 +585,30 @@ alloc_ext_pp()
 		ext_pp = p;
 	}
 
-	size = 0;
-	mu_list_foreach(include_path, _count_include_size, &size);
-	if (size) {
-		ext_pp = mu_realloc(ext_pp, strlen(ext_pp) + size + 1);
-		mu_list_foreach(include_path, _append_includes, ext_pp);
+	if (preprocessor.pass_includes) {
+		size = 0;
+		path_enumerate(include_path.base, _count_include_size, &size);
+		if (size) {
+			ext_pp = mu_realloc(ext_pp, strlen(ext_pp) + size + 1);
+			path_enumerate(include_path.base, _append_includes, ext_pp);
+		}
 	}
+	
+	setup_file = preprocessor.setup_file;
 
-	if (try_file("pp-setup", 1, 0, &setup_file)) {
+	if (setup_file[0] == '/' ||
+	    try_file(setup_file, NULL, &include_path, 1, 0, &setup_file)) {
 		ext_pp = mu_realloc(ext_pp, strlen(ext_pp) +
-				  strlen(setup_file) + 2);
+				    strlen(setup_file) + 2);
 		strcat(ext_pp, " ");
 		strcat(ext_pp, setup_file);
 	}
+
+	preprocessor.command = ext_pp;
 }
 
 int
-preprocess_input()
+preprocess_input(void)
 {
 	char buffer[512];
 	FILE *file;
@@ -512,4 +623,3 @@ preprocess_input()
 	pp_extrn_shutdown(file, pid);
 	return 0;
 }
-
