@@ -634,37 +634,48 @@ prog_trace(eval_environ_t env, const char *fmt, ...)
 void instr_builtin(eval_environ_t env);
 void instr_funcall(eval_environ_t env);
 void instr_locus(eval_environ_t env);
+void instr_catch(eval_environ_t env);
+
+static struct mu_locus_point *
+stack_find_locus(eval_environ_t env, prog_counter_t pc, prog_counter_t off, struct mu_locus_point *pt)
+{
+	int i;
+
+	for (i = 0; i < 32; i++, off++) {
+		if (pc > off && prog[pc - off].c_instr == instr_locus) {
+			pt->mu_file = (char*)(env->dataseg
+					      + mf_cell_c_value(prog[pc - off + 1], size));
+			pt->mu_line = mf_cell_c_value(prog[pc - off + 2], size);
+			return pt;
+		}
+	}
+	return NULL;
+}
 
 static void
 stack_trace_entry(eval_environ_t env, prog_counter_t pc)
 {
-	int i;
 	char *name;
-	struct mu_locus_point *ploc = NULL, loc;
+	struct mu_locus_point *pt, ptbuf;
 	
 	if (mf_cell_instr(prog[pc-2]) == instr_funcall ||
 	    mf_cell_instr(prog[pc-2]) == instr_builtin) {
 		name = (char*)(env->dataseg + mf_cell_c_value(prog[pc-1], size));
 		pc -= 2;
-	} else {
+		pt = stack_find_locus(env, pc, 3, &ptbuf);
+	} else if (mf_cell_instr(prog[pc-2]) == instr_catch) {
 		name = "(in catch)";
-		pc -= 3;
+		pc -= 2;
+		pt = stack_find_locus(env, pc, 3, &ptbuf);
+	} else {
+		name = "(unknown)";
+		pt = stack_find_locus(env, pc, 0, &ptbuf);
 	}
-	
-	for (i = 0; i < 10; i++) {
-		if (pc > i + 3 && prog[pc - i - 3].c_instr == instr_locus) {
-			loc.mu_file = (char*)(env->dataseg
-					      + mf_cell_c_value(prog[pc - i - 2], size));
-			loc.mu_line = mf_cell_c_value(prog[pc - i - 1], size);
-			ploc = &loc;
-			break;
-		}
-	}
-		
-	if (ploc)
+
+	if (pt)
 		mu_error("%04lu: %s:%u: %s",
 			 (unsigned long) pc,
-			 ploc->mu_file, ploc->mu_line, name);
+			 pt->mu_file, pt->mu_line, name);
 	else
 		mu_error("%04lu: %s",
 			 (unsigned long) pc, name);
@@ -679,10 +690,13 @@ runtime_stack_trace(eval_environ_t env)
 	stack_trace_entry(env, env->pc);
 	for (base = env->base; base < datasize + env->stack_size - 4;
 	     base = mf_c_val(env->dataseg[base + 1], size) + base + 1) {
-		prog_counter_t pc = mf_c_val(env->dataseg[base + 2], ulong) - 1;
+		prog_counter_t pc = mf_c_val(env->dataseg[base + 2], ulong);
 		if (pc < 2)
 			break; /*FIXME*/
 		stack_trace_entry(env, pc);
+		if (mf_cell_instr(prog[pc-2]) == instr_catch) {
+			stack_trace_entry(env, mf_c_val(env->dataseg[base + 5], size));
+		}
 	}
 	mu_error(_("stack trace finishes"));
 }
@@ -1001,9 +1015,42 @@ instr_xchg(eval_environ_t env)
 void
 instr_dup(eval_environ_t env)
 {
-	if (PROG_TRACE_ENGINE)
+        if (PROG_TRACE_ENGINE)
 		prog_trace(env, "DUP");
-	push(env, get_arg(env, 0));
+	push(env, (STKVAL)0L);
+	env->dataseg[env->tos+1] = env->dataseg[env->tos+2];
+}
+
+/*
+ * push COUNT
+ * push OFFSET
+ * dupn FRAME
+ *
+ * Pushes on stack COUNT stack words starting from OFFSET in the given
+ * FRAME, in the same order they appear on stack.
+ */
+void
+instr_dupn(eval_environ_t env)
+{
+	size_t frame = mf_c_val(get_immediate(env, 0), size);
+	long count = mf_c_val(get_arg(env, 1), long);
+	long off = mf_c_val(get_arg(env, 0), long);
+	size_t i;
+
+	if (count < 0)
+		runtime_error(env, _("dupn: negative count"));
+	if (PROG_TRACE_ENGINE)
+		prog_trace(env, "DUPN frame=%zu off=%ld count=%zu",
+			   frame, off, count);
+
+	adjust_stack(env, 2);
+
+	for (i = count; i > 0; i--) {
+		push(env, (STKVAL)0L);
+		env->dataseg[env->tos+1] = env->dataseg[env_base(env, frame) + off + i - 1];
+	}
+	
+	advance_pc(env, 1);
 }
 
 void
@@ -1042,7 +1089,7 @@ instr_memstk(eval_environ_t env)
 	      push(env, env_base(env, frame) + off)
 
 	   However, push() itself might cause dataseg expansion,
-	   which would change tos/toh pointers and render the previously
+	   which would change tos/base pointers and render the previously
 	   computed argument value invalid.  The correct way is to push
 	   a placeholder value first, thereby occupying stack slot and
 	   eventually causing dataseg expansion, and then compute val
@@ -1254,7 +1301,7 @@ dump_symbol(prog_counter_t i)
 	       (char *) (dataseg + mf_cell_c_value(prog[i], size)));
 }
 
-/* Comparation instructions */
+/* Comparison instructions */
 void
 instr_eqn(eval_environ_t env)
 {
@@ -1444,7 +1491,7 @@ instr_jmp(eval_environ_t env)
 	advance_pc(env, off+1);
 }
 
-/* Longean instructions */
+/* Boolean instructions */
 void
 instr_not(eval_environ_t env)
 {
@@ -1455,7 +1502,7 @@ instr_not(eval_environ_t env)
 	pushn(env, !v);
 }
 
-/* Bitwise arithmetics */
+/* Bitwise arithmetic */
 void
 instr_logand(eval_environ_t env)
 {
@@ -1985,12 +2032,6 @@ instr_concat(eval_environ_t env)
 	push(env, (STKVAL) off);
 }
 
-void
-dump_adjust(prog_counter_t i)
-{
-	printf("%lu ", mf_cell_c_value(prog[i], ulong));
-}
-
 void
 instr_asgn(eval_environ_t env)
 {
@@ -2053,7 +2094,6 @@ instr_throw(eval_environ_t env)
 {
 	unsigned long n = mf_c_val(get_immediate(env, 0), ulong);
 	size_t off = mf_c_val(get_arg(env, 0), size);
-	advance_pc(env, 1);
 	adjust_stack(env, 1);
 	if (n > exception_count)
 		runtime_error(env, _("invalid exception number: %lu"), n);
@@ -2088,7 +2128,6 @@ instr_return(eval_environ_t env)
 	if (PROG_TRACE_ENGINE)
 		prog_trace(env, "RETURN");
 	env_leave_frame(env, 0);
-	env->pc--;
 }
 
 void
@@ -2097,16 +2136,25 @@ instr_retcatch(eval_environ_t env)
 	prog_counter_t pc = env->pc;
 	if (PROG_TRACE_ENGINE)
 		prog_trace(env, "RETCATCH");
-	env_leave_frame(env, 2);
+	env_leave_frame(env, 3);
 	env->pc = pc;
 }
 
+/*
+ * Return true if the given bit is set in the exception mask starting at
+ * offset off.
+ */
+static inline int
+env_exmask_isset(eval_environ_t env, size_t off, int bit)
+{
+	return ((bitmask_bits_t)mf_c_val(env->dataseg[off + 1 + BIT_WORD(bit)], uint)) & BIT_MASK(bit);
+}
+
 void
 instr_saveex(eval_environ_t env)
 {
 	size_t off = mf_c_val(get_immediate(env, 0), size);
 	size_t count = mf_c_val(env->dataseg[off], size);
-	STKVAL *tab = (STKVAL *) (env->dataseg + off + 1);
 	size_t i;
 
 	if (PROG_TRACE_ENGINE)
@@ -2115,8 +2163,13 @@ instr_saveex(eval_environ_t env)
 			   (unsigned long) count);
 
 	advance_pc(env, 1);
+	/*
+	 * Note: in the loop that follows each push can lead to data segment
+	 * expansion, therefore the actual start address of the exception table
+	 * needs to be computed in each iteration using env_exmask_isset.
+	 */
 	for (i = 0; i < count * NBMBITS; i++)
-		if (((bitmask_bits_t)mf_c_val(tab[BIT_WORD(i)], uint)) & BIT_MASK(i)) {
+		if (env_exmask_isset(env, off, i)) {
 			mu_debug(MF_SOURCE_PROG, MU_DEBUG_TRACE9,
 			      ("Push Exception: %lu %lu <- pc=%lu, tos=%lu, base=%lu",
 			       (unsigned long) i,
@@ -2176,6 +2229,12 @@ instr_restex(eval_environ_t env)
 	} while (i > 0);
 }
 
+void
+dump_adjust(prog_counter_t i)
+{
+	printf("%lu ", mf_cell_c_value(prog[i], ulong));
+}
+
 void
 instr_adjust(eval_environ_t env)
 {
@@ -2186,6 +2245,30 @@ instr_adjust(eval_environ_t env)
 	advance_pc(env, 1);
 }
 
+/*
+ * Adjust stack after return from a variadic function.
+ *
+ * When such function is called, it is passed the actual number
+ * of arguments in its first (hidden) argument.
+ *
+ * The inline parameter to adjustx gives the number of arguments the
+ * function is declared to take (both mandatory and optional).  These
+ * are always on stack, so this value is used instead of the actual
+ * number if the latter is less than the former.  This can happen,
+ * e.g. if a function takes both optional and variable arguments and
+ * is given only mandatory ones.
+ */
+void
+instr_adjustx(eval_environ_t env)
+{
+	long nargs = mf_c_val(get_immediate(env, 0), long);
+	size_t n = mf_c_val(pop(env), size);
+	if (PROG_TRACE_ENGINE)
+		prog_trace(env, "ADJUSTX %zu %ld", n, nargs);
+	adjust_stack(env, n < nargs ? nargs : n);
+	advance_pc(env, 1);
+}
+
 void
 instr_popreg(eval_environ_t env)
 {
@@ -2417,18 +2500,10 @@ env_init_dataseg(eval_environ_t env)
 	init_dataseg(env->dataseg, dvarsize);
 }
 
-void
-env_make_frame0(eval_environ_t env)
-{
-	push(env, (STKVAL) 0);
-	push(env, (STKVAL) (env->base - env->tos));
-	env->base = env->tos;
-}
-
 void
 env_make_frame(eval_environ_t env)
 {
-	push(env, (STKVAL) (env->pc + 1));
+	push(env, (STKVAL) env->pc);
 	push(env, (STKVAL) (env->base - env->tos));
 	env->base = env->tos;
 }
@@ -2550,6 +2625,8 @@ env_throw_0(eval_environ_t env, mf_exception status, size_t off)
 	
 	pc = env->catch_ctx[status].pc;
 	if (pc) {
+		prog_counter_t throw_pc = env->pc;
+
 		/* Restore tos */
 		env->tos = TOS_INVARIANT(env, env->catch_ctx[status].tos);
 		env->base = TOS_INVARIANT(env, env->catch_ctx[status].base);
@@ -2558,6 +2635,7 @@ env_throw_0(eval_environ_t env, mf_exception status, size_t off)
 		/* Fixup the program counter */
 		env->pc = pc - 1;
 		/* Generate normal entry frame */
+		push(env, (STKVAL) throw_pc);
 		push(env, (STKVAL) off); 
 		env_push_number(env, status);
 		env_make_frame(env);
@@ -2635,7 +2713,7 @@ env_capture_start(eval_environ_t env)
 		/* Drop any previously captured message data */
 		env_free_captured(env);
 		mu_header_destroy(&env->header);
-		/* Truncate existing stream and reposition to its
+		/* Truncate existing stream and re-position to its
 		   beginning */
 		rc = mu_stream_truncate(env->stream, 0);
 		if (rc == 0 &&
@@ -2777,7 +2855,7 @@ env_get_header(eval_environ_t env, mu_header_t *hdr)
 		}
 		text[total] = 0;
 
-               /* FIXME: Reposition the stream at its end.
+               /* FIXME: Re-position the stream at its end.
 		  This call may happen in the middle of capturing
 		  so I have to make sure to not disturb the capturing
 		  process.
@@ -3025,6 +3103,63 @@ destroy_environment(eval_environ_t env)
 	env_builtin_priv_destroy(env);
 	free(env);
 }
+
+static const char *
+null_getsym(void *data, const char *str)
+{
+	eval_environ_t *env = data;
+	runtime_error(*env, "%s", _("invalid use of $macro in startup handler"));
+}
+
+static int
+null_setreply(void *data, char *code, char *xcode, char *message)
+{
+	eval_environ_t *env = data;
+	runtime_error(*env, "%s", _("invalid use of milter action in startup handler"));
+}
+static void
+null_msgmod(void *data, struct msgmod_closure *clos)
+{
+	eval_environ_t *env = data;
+	runtime_error(*env, "%s", _("invalid use of message modification function in startup handler"));
+}
+
+void
+run_shutdown(void)
+{
+	if (entry_point[smtp_state_shutdown]) {
+		eval_environ_t env;
+		env = create_environment(NULL,
+					 null_getsym,
+					 null_setreply,
+					 null_msgmod,
+					 &env);
+		env_init(env);
+		xeval(env, smtp_state_shutdown);
+		destroy_environment(env);
+	}
+	mu_stream_close(mf_strecho);
+}
+
+void
+run_startup(void)
+{
+	if (entry_point[smtp_state_startup]) {
+		eval_environ_t env;
+		env = create_environment(NULL,
+					 null_getsym,
+					 null_setreply,
+					 null_msgmod,
+					 &env);
+		env_init(env);
+		if (xeval(env, smtp_state_startup))
+			exit(EX_CONFIG);
+		env_final_gc(env);
+		datasize = env->toh;
+		memcpy(dataseg, env->dataseg, datasize * sizeof dataseg[0]);
+		destroy_environment(env);
+	}
+}
 
 
 struct entry_point {
@@ -3120,7 +3255,7 @@ print_code()
 				 function_counter, &d);
 	}
 	epcount = d.i;
-	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
+	for (tag = smtp_state_startup; tag < smtp_state_count; tag++)
 		if (entry_point[tag])
 			epcount++;
 	
@@ -3128,7 +3263,7 @@ print_code()
 
         /* Fill in entry points array */
 	i = 0;
-	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
+	for (tag = smtp_state_startup; tag < smtp_state_count; tag++)
 		if (entry_point[tag]) {
 			ep[i].ishandler = 1;
 			ep[i].pc = entry_point[tag];
@@ -3213,8 +3348,9 @@ env_final_gc(eval_environ_t env)
 			env->dataseg[off] = (STKVAL) q;
 #undef S_PTR
 		}
-		free(s_off);
 	}
+
+	free(s_off);
 }
 
 void
