@@ -74,6 +74,8 @@ prog_counter_t code_exmask(struct exmask *exmask);
 prog_counter_t code_get_counter();
 void dump_code(prog_counter_t start, prog_counter_t end);
 void fixup_code(void);
+void run_startup(void);
+void run_shutdown(void);
 
 /* Runtime evaluation functions */
 struct eval_environ;
