@@ -261,6 +261,31 @@ getval(char *input)
 	return p;
 }
 
+static struct mu_kwd tls_kwd[] = {
+	{ "never",      tls_never },
+	{ "always",	tls_always },
+	{ "ondemand",   tls_ondemand },
+	{ NULL }
+};
+
+int
+smtp_str_to_starttls(char const *str, smtp_starttls_t *val)
+{
+	int v;
+	if (mu_kwd_xlat_name(tls_kwd, str, &v))
+		return -1;
+	*val = v;
+	return 0;
+}
+
+char const *
+smtp_starttls_to_str(smtp_starttls_t t)
+{
+	char const *val = NULL;
+	mu_kwd_xlat_tok(tls_kwd, t, &val);
+	return val;
+}
+
 static struct vrfy_queue *
 addq(int argc, char **argv, const char **errp)
 {
@@ -271,11 +296,22 @@ addq(int argc, char **argv, const char **errp)
 	char *ehlo = ehlo_domain;
 	char *mailfrom = mailfrom_address;
 	enum callout_mode mode = cmode_mx_first;
-	
+	smtp_starttls_t tls = smtp_use_tls;
+
+	static struct mu_kwd mode_kwd[] = {
+		{ "mxfirst",    cmode_mx_first },
+		{ "default",    cmode_mx_first },
+		{ "mxonly",     cmode_mx_only },
+		{ "hostonly",   cmode_host_only },
+		{ "hostfirst",  cmode_host_first },
+		{ NULL }
+	};
+
 	email = argv[1];
 	for (i = 2; i < argc; i++) {
 		char *attr = argv[i];
 		char *val = getval(attr);
+		int v;
 		
 		if (!val) {
 			*errp = "syntax error";
@@ -284,18 +320,21 @@ addq(int argc, char **argv, const char **errp)
 		if (!strcasecmp(attr, "host")) {
 			host = val;
 		} else if (!strcasecmp(attr, "mode")) {
-			if (!strcmp(val, "mxfirst") || !strcmp(val, "default"))
-				mode = cmode_mx_first;
-			else if (!strcmp(val, "mxonly"))
-				mode = cmode_mx_only;
-			else if (!strcmp(val, "hostonly"))
-				mode = cmode_host_only;
-			else if (!strcmp(val, "hostfirst"))
-				mode = cmode_host_first;
+			if (mu_kwd_xlat_name(mode_kwd, val, &v)) {
+				*errp = "unrecognized value for mode";
+				return NULL;
+			}
+			mode = v;
 		} else if (!strcasecmp(attr, "ehlo")) {
 			ehlo = val;
 		} else if (!strcasecmp(attr, "mailfrom")) {
 			mailfrom = val;
+		} else if (!strcasecmp(attr, "starttls")) {
+			if (smtp_str_to_starttls(val, &tls)) {
+				*errp = "unrecognized value for tls";
+				return NULL;
+			}
+			tls = v;
 		} else {
 			*errp = "unknown attribute";
 			return NULL;
@@ -321,7 +360,7 @@ addq(int argc, char **argv, const char **errp)
 				smtp_timeout_hard,
 				savsrv_smtp_io_callback,
 				qp);
-	smtp_io_setup_callout(qp->io, email, ehlo, mailfrom);
+	smtp_io_setup_callout(qp->io, email, ehlo, mailfrom, tls);
 	switch (mode) {
 	case cmode_host_first:
 	case cmode_host_only:
@@ -362,14 +401,16 @@ cmd_get(FILE *fp, int argc, char **argv)
 
 	for (i = 1; i < argc; i++) {
 		char *attr = argv[i];
-		char *val;
+		char const *val;
 		if (!mu_c_strcasecmp(attr, "ehlo"))
 			val = ehlo_domain;
 		else if (!mu_c_strcasecmp(attr, "mailfrom"))
 			val = mailfrom_address;
 		else if (!mu_c_strcasecmp(attr, "transcript"))
 			val = smtp_transcript ? "ON" : "OFF";
-		else /* FIXME: Timeouts */
+		else if (!mu_c_strcasecmp(attr, "starttls")) {
+			val = smtp_starttls_to_str(smtp_use_tls);
+		} else /* FIXME: Timeouts */
 			val = NULL;
 		if (val)
 			writeout(fp, "* %s=%s\r\n", attr, val);
@@ -382,6 +423,7 @@ int
 cmd_set(FILE *fp, int argc, char **argv)
 {
 	int i;
+	smtp_starttls_t tls;
 
 	for (i = 1; i < argc; i++) {
 		char *val = strchr(argv[i], '=');
@@ -399,6 +441,12 @@ cmd_set(FILE *fp, int argc, char **argv)
 				writeout(fp, "NO bad assignment %s\n", argv[i]);
 				return 1;
 			}
+		} else if (!mu_c_strncasecmp(argv[i], "starttls", 3)) {
+			val++;
+			if (smtp_str_to_starttls(val, &tls)) {
+				writeout(fp, "NO unrecognized starttls mode: %s\n", val);
+				return 1;
+			}
 		} else {
 			writeout(fp, "NO bad assignment %s\n", argv[i]);
 			return 1;
@@ -415,8 +463,11 @@ cmd_set(FILE *fp, int argc, char **argv)
 		} else if (!mu_c_strncasecmp(attr, "mailfrom", 8)) {
 			free(mailfrom_address);
 			mailfrom_address = mu_strdup(val);
-		} else if (!mu_c_strncasecmp(attr, "transcript", 10))
+		} else if (!mu_c_strncasecmp(attr, "transcript", 10)) {
 			smtp_transcript = mu_c_strcasecmp(val, "ON") == 0;
+		} else if (!mu_c_strncasecmp(attr, "starttls", 3)) {
+			smtp_use_tls = tls;
+		}
 	}
 	writeout(fp, "OK\r\n");
 	return 0;
