@@ -500,7 +500,7 @@ struct spf_term {
 	spf_term_type type;           /* Term type */
 	const char *expr;             /* Original expression (for debugging)*/ 
 	spf_term_handler handler;     /* Term handler */
-	int has_arg;                  /* Is an explicite argument given */
+	int has_arg;                  /* Is an explicit argument given */
 	spf_term_arg arg;             /* Argument if has_arg==1 */
 	spf_result qualifier;         /* only for type == spf_term_directive */
 	unsigned long masklen;        /* Netmask length */
@@ -529,7 +529,7 @@ struct spf_term_syntax {
 	}		
 
 
-/* Hanlders for particular terms */
+/* Handlers for particular terms */
 
 /* RFC 4408, 5.1.
      all              = "all"
@@ -641,7 +641,7 @@ mech_a(spf_data *dat, spf_term_arg *arg, unsigned long masklen)
 
 	res = spf_term_nomatch;
 	for (i = 0; i < r.count; i++) {
-		if (ntohl(r.data.ip[i] & netmask) == addr.s_addr) {
+		if ((ntohl(r.data.ip[i]) & netmask) == addr.s_addr) {
 			res = spf_term_match;
 			break;
 		}
