@@ -37,6 +37,12 @@ int server_flags = 0;
 char *log_stream = DEFAULT_LOG_STREAM;
 char *pidfile;                
 int smtp_transcript;
+smtp_starttls_t smtp_use_tls = tls_ondemand;
+struct mu_tls_config tls_conf = {
+	.priorities = "NORMAL:%COMPAT",
+	.handshake_timeout = 60
+};
+
 static int transcript_option;
 struct mu_sockaddr *source_address;  /* Source address for TCP connections */
 int mtasim_option;     /* mtasim compatibility mode */
@@ -472,6 +478,14 @@ cb_state_directory(void *data, mu_config_value_t *arg)
 	return mf_option_state_directory(arg->v.string);
 }
 
+static int
+cb_starttls(void *data, mu_config_value_t *arg)
+{
+	if (mu_cfg_assert_value_type(arg, MU_CFG_STRING))
+		return 1;
+	return smtp_str_to_starttls(arg->v.string, (smtp_starttls_t*)data);
+}
+
 /* FIXME: Umask not configurable */
 
 static struct mu_cfg_param srv_cfg_param[] = {
@@ -487,6 +501,12 @@ static struct mu_cfg_param srv_cfg_param[] = {
 	  N_("Enable transcript of call-out SMTP sessions.") },
 
 	{ "smtp-timeout", mu_cfg_section, },
+	{ "smtp-starttls", mu_cfg_callback, &smtp_use_tls, 0, cb_starttls,
+	  N_("Use of STARTTLS in callout sessions (default: ondemand)."),
+	  N_("arg: never|always|ondemand") },
+
+	{ "tls", mu_cfg_section, &tls_conf },
+	
         { "io-timeout", mu_cfg_callback, &io_timeout, 0, cb_timeout,
 	  N_("Timeout for all SMTP I/O operations."),
           N_("time: seconds") },
@@ -699,6 +719,7 @@ mf_srvcfg_init(const char *progname, const char *label)
 
 	smtp_timeout_cfg_init();
 	server_stmt_init(label);
+	mu_tls_cfg_init ();
 
 	mailfromd_state_dir = DEFAULT_STATE_DIR;
 	if (!pidfile)
