@@ -842,7 +842,7 @@ function_lookup_or_install(const char *name, struct mu_locus_range const *locus)
 
 struct function *
 function_install(const char *name, size_t parmcnt, size_t optcnt, int varargs,
-		 data_type_t *parmtypes, data_type_t rettype,
+		 data_type_t vartype, data_type_t *parmtypes, data_type_t rettype,
 		 struct mu_locus_range const *locus)
 {
 	struct function *fp = function_lookup_or_install(name, locus);
@@ -854,6 +854,7 @@ function_install(const char *name, size_t parmcnt, size_t optcnt, int varargs,
 	fp->parmcount = parmcnt;
 	fp->optcount = optcnt;
 	fp->varargs = varargs;
+	fp->vartype = vartype;
 	fp->parmtype = parmtypes;
 	fp->rettype = rettype;
 	fp->exmask = exmask_create();
